import 'dart:convert';

import 'package:contentsquare/contentsquare.dart';
import 'package:contentsquare_platform_interface/contentsquare_platform_interface.dart';
import 'package:fast_immutable_collections/fast_immutable_collections.dart';
import 'package:flutter/services.dart';
import 'package:flutter_test/flutter_test.dart';

void main() {
  setUpAll(() {
    TestWidgetsFlutterBinding.ensureInitialized();
  });

  group(
    '$Contentsquare',
    () {
      const channel = MethodChannel('contentsquare');
      late IList<MethodCall> log;
      late Contentsquare contentsquare;

      setUp(() {
        log = IList<MethodCall>();
        contentsquare = Contentsquare();

        TestDefaultBinaryMessengerBinding.instance!.defaultBinaryMessenger
            .setMockMethodCallHandler(
          channel,
          (MethodCall methodCall) async {
            log = log.add(methodCall);
            if (methodCall.method == 'getUserId') {
              return 'Test User Id';
            }
            return null;
          },
        );
      });

      test('send', () async {
        await contentsquare.send('Test screen');
        expect(log, <Matcher>[
          isMethodCall('send', arguments: {'screenName': 'Test screen'}),
        ]);
      });

      test(
        'sendTransaction',
        () async {
          await contentsquare.sendTransaction(42, 'EUR', 'testId');
          final expectedMatcher = isMethodCall(
            'sendTransaction',
            arguments: {
              'price': 42,
              'currency': 'EUR',
              'transactionId': 'testId'
            },
          );
          expect(log, <Matcher>[expectedMatcher]);
        },
      );

      test('sendDynamicVar', () async {
        await contentsquare.sendDynamicVar('key', intValue: 42);
        await contentsquare.sendDynamicVar('key', stringValue: 'test');

        final expectedMatcher0 = isMethodCall(
          'sendDynamicVar',
          arguments: {
            'key': 'key',
            'stringValue': null,
            'intValue': 42,
          },
        );
        final expectedMatcher1 = isMethodCall(
          'sendDynamicVar',
          arguments: {
            'key': 'key',
            'stringValue': 'test',
            'intValue': null,
          },
        );
        expect(log, equals([expectedMatcher0, expectedMatcher1]));
      });

      test(
        'optIn',
        () async {
          await contentsquare.optIn();

          expect(log, <Matcher>[isMethodCall('optIn', arguments: null)]);
        },
      );

      test(
        'optOut',
        () async {
          await contentsquare.optOut();

          expect(log, <Matcher>[isMethodCall('optOut', arguments: null)]);
        },
      );

      test(
        'forgetMe',
        () async {
          await contentsquare.forgetMe();

          expect(log, <Matcher>[isMethodCall('forgetMe', arguments: null)]);
        },
      );

      test(
        'getUserId',
        () async {
          final userId = await contentsquare.getUserId();

          expect(log, <Matcher>[isMethodCall('getUserId', arguments: null)]);

          expect(userId, 'Test User Id');
        },
      );

      test(
        'stopTracking',
        () async {
          await contentsquare.stopTracking();

          expect(log, <Matcher>[isMethodCall('stopTracking', arguments: null)]);
        },
      );

      test(
        'resumeTracking',
        () async {
          await contentsquare.resumeTracking();

          expect(
            log,
            <Matcher>[isMethodCall('resumeTracking', arguments: null)],
          );
        },
      );

      test(
        'sendTouchEvent',
        () async {
          await contentsquare.sendTouchEvent(
            version: 1,
            type: ContentsquareGesture.longPress,
            data: const TouchEventData(path: 'path/to/my/widget'),
          );

          final expectedMethodCall = isMethodCall(
            'sendTouchEvent',
            arguments: {
              'body': jsonEncode({
                'type': 8,
                'version': 1,
                'data': {
                  'path': 'path/to/my/widget',
                  'distance': null,
                  'direction': null,
                  'velocity': null
                },
              }),
            },
          );

          expect(log, <Matcher>[expectedMethodCall]);
        },
      );
    },
  );
}
