import 'dart:math';

import 'package:contentsquare/contentsquare.dart';
import 'package:contentsquare/src/modules/screengraph/hit_test/get_tvp_at_hit_position.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter_test/flutter_test.dart';

import '../../../helpers/decompress_tvp.dart';
import '../../../helpers/expect_logs.dart';
import '../../../helpers/matchers/fixtures/matcher.dart';
import '../../../helpers/get_center.dart';
import '../../../helpers/type_of.dart';

class OutOfBoundsHit extends SingleChildRenderObjectWidget {
  const OutOfBoundsHit({Key? key, required this.size}) : super(key: key);

  final Size size;

  @override
  RenderObject createRenderObject(BuildContext context) =>
      RenderOutOfBoundsHit(customSize: size);
}

class RenderOutOfBoundsHit extends RenderBox {
  RenderOutOfBoundsHit({required this.customSize});

  final Size customSize;

  @override
  void performLayout() {
    size = Size(
      min(customSize.width, constraints.maxWidth),
      min(customSize.height, constraints.maxHeight),
    );
  }

  @override
  bool hitTest(BoxHitTestResult result, {required Offset position}) {
    result.add(BoxHitTestEntry(this, position));
    return true;
  }
}

class _NamedWidget<Name> extends StatelessWidget {
  const _NamedWidget({Key? key, required this.child}) : super(key: key);

  final Widget child;

  @override
  Widget build(BuildContext context) => child;
}

class _Included {}

void main() {
  group(
    'getTVPAtHitPosition',
    () {
      testWidgets(
        'Should return the TVP of the last hittable widget around this '
        'position and ignore the inner Text which is not hittable',
        (WidgetTester tester) async {
          const key = ValueKey('Text');
          await tester.pumpWidget(
            MaterialApp(
              home: Scaffold(
                // [GestureDetector] will capture the tap but the TVP should
                // extends to the [SizedBox].
                body: GestureDetector(
                  child: const Center(
                    child: SizedBox(
                      width: 100,
                      height: 100,
                      // Text is not hittable.
                      child: Text(
                        'Text',
                        key: key,
                      ),
                    ),
                  ),
                ),
              ),
            ),
          );

          final hitTVPs = getTVPAtHitPosition(
            getCenterFromRenderBoxKey(key: key),
          );
          final decompressedHitTVPs = hitTVPs
              .map(
                (hitTVP) => decompressTVP(tvp: hitTVP),
              )
              .toList();

          expect(
            decompressedHitTVPs.single,
            matchesFixture('hitTVPNotHittable'),
          );
        },
      );

      group(
        'Scrollables',
        () {
          testWidgets(
            'Should return the right TVP '
            'when the widget tapped is in a scrollable',
            (WidgetTester tester) async {
              await tester.pumpWidget(
                MaterialApp(
                  home: ListView.builder(
                    itemBuilder: (_, index) => SizedBox(
                      height: 50,
                      child: Text(
                        'Text $index',
                        key: ValueKey('Text $index'),
                      ),
                    ),
                    itemCount: 2000,
                  ),
                ),
              );

              final hitTVPs = getTVPAtHitPosition(
                getCenterFromRenderBoxKey(key: const ValueKey('Text 3')),
              );
              final decompressedHitTVPs =
                  hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

              expect(
                decompressedHitTVPs.single,
                matchesFixture('hitTVPWithinScrollable'),
              );
            },
          );

          testWidgets(
            'Should return the right TVP when the widget tapped is in a '
            'scrollable and the scrollable needs to be scrolled to reveal '
            'the widget',
            (WidgetTester tester) async {
              await tester.pumpWidget(
                MaterialApp(
                  home: ListView.builder(
                    itemBuilder: (context, index) => SizedBox(
                      height: MediaQuery.of(context).size.height / 10,
                      child: Text(
                        'Text $index',
                        key: ValueKey('Text $index'),
                      ),
                    ),
                    itemCount: 30,
                  ),
                ),
              );

              const itemKey = ValueKey('Text 20');
              final itemFinder = find.byKey(itemKey);
              await tester.scrollUntilVisible(itemFinder, 100);

              final hitTVPs = getTVPAtHitPosition(
                getCenterFromRenderBoxKey(key: itemKey),
              );
              final decompressedHitTVPs =
                  hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

              expect(
                decompressedHitTVPs.single,
                matchesFixture('hitTVPNeedsScroll'),
              );
            },
          );
        },
      );

      testWidgets(
        'Should return one TVP of the parent which is in the screengraph '
        'when several widgets have the same parent in the screengraph and '
        'are not themselves in the screengraph\n'
        'This tests the "removeDuplicate" part of the algorithm',
        (WidgetTester tester) async {
          const key = ValueKey('Stack');
          await tester.pumpWidget(
            MaterialApp(
              home: Scaffold(
                body: SizedBox(
                  width: 100,
                  height: 100,
                  child: Stack(
                    key: key,
                    fit: StackFit.expand,
                    children: const [
                      // The two widget are placed at the bottom of the stack.
                      // The center of the stack will be hit, therefore the
                      // Align will be reported with the same TVP (the
                      // one ending at the Stack).
                      Align(
                        alignment: Alignment.bottomCenter,
                        child: Text('Text 1'),
                      ),
                      Align(
                        alignment: Alignment.bottomCenter,
                        child: Text('Text 2'),
                      ),
                    ],
                  ),
                ),
              ),
            ),
          );

          final hitTVPs = getTVPAtHitPosition(
            getCenterFromRenderBoxKey(key: key),
          );
          final decompressedHitTVPs =
              hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

          expect(
            decompressedHitTVPs.single,
            matchesFixture('hitTVPRemoveDuplicates'),
          );
        },
      );

      testWidgets(
        'Should return one TVP when a matched TVP are included in another '
        'when searching for widgets at the hit position\n'
        'This tests the "removeInclusion" part of the algorithm',
        (WidgetTester tester) async {
          const key = ValueKey('Stack');
          await tester.pumpWidget(
            MaterialApp(
              home: Scaffold(
                body: SizedBox(
                  width: 100,
                  height: 100,
                  child: Stack(
                    key: key,
                    fit: StackFit.expand,
                    children: const [
                      // The text won't be in the TVP because the middle of the
                      // Stack is hit. Therefore the TVP will end in the Stack,
                      // which is included in the TVP of the "SizedBox".
                      Align(
                        alignment: Alignment.bottomCenter,
                        child: Text('Text 1'),
                      ),

                      // The "SizedBox" widget will be hit and form a TVP
                      // including the one ending at the Stack.
                      Center(
                        child: SizedBox(height: 50, width: 50),
                      ),
                    ],
                  ),
                ),
              ),
            ),
          );

          final hitTVPs = getTVPAtHitPosition(
            getCenterFromRenderBoxKey(key: key),
          );
          final decompressedHitTVPs =
              hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

          expect(
            decompressedHitTVPs.single,
            matchesFixture('hitTVPRemoveInclusion'),
          );
        },
      );

      testWidgets(
        'Should return several TVPs when we cannot differentiate between them '
        'because the hit target returned by the hit test is a parent '
        'of both',
        (WidgetTester tester) async {
          const key = ValueKey('Stack');
          await tester.pumpWidget(
            MaterialApp(
              home: Scaffold(
                body: SizedBox(
                  width: 100,
                  height: 100,
                  // The Stack will respond the the hitTest but the [SizedBox]s
                  // will not.
                  child: Stack(
                    key: key,
                    fit: StackFit.expand,
                    children: const [
                      // This is in the tapped zone and therefore will be
                      // included
                      Center(
                        child: SizedBox(height: 50, width: 50),
                      ),

                      // This is in the tapped zone and therefore will be
                      // included
                      Center(
                        child: SizedBox(height: 50, width: 50),
                      ),
                    ],
                  ),
                ),
              ),
            ),
          );

          final hitTVPs = getTVPAtHitPosition(
            getCenterFromRenderBoxKey(key: key),
          );
          final decompressedHitTVPs = hitTVPs
              .map(
                (hitTVP) => decompressTVP(tvp: hitTVP),
              )
              .toList();

          // The order does not matter, so we don't use "listEquals".
          expect(hitTVPs.length, 2);
          expect(
            decompressedHitTVPs[0],
            matchesFixture('multipleHitTVPs0'),
          );
          expect(
            decompressedHitTVPs[1],
            matchesFixture('multipleHitTVPs1'),
          );
        },
      );

      testWidgets(
        'Should return an empty list and emit a log '
        'when a [RenderObject] responds to a hit which is not it its globalFrame',
        (WidgetTester tester) async {
          const key = ValueKey('App');
          await tester.pumpWidget(
            MaterialApp(
              key: key,
              home: Scaffold(
                body: GestureDetector(
                  child: const Align(
                    alignment: Alignment.bottomCenter,
                    child: OutOfBoundsHit(
                      size: Size(double.infinity, 10),
                    ),
                  ),
                ),
              ),
            ),
          );

          expectLogs((log) => expect(log, isA<HitElementNotFoundLog>()));

          final hitTVPs = getTVPAtHitPosition(
            getCenterFromRenderBoxKey(key: key),
          );

          expect(hitTVPs, isEmpty);
        },
      );

      group('Gestures inside or at [ContentsquareWebViewTrackerBuilder]', () {
        testWidgets(
          'should not return any TVPs '
          'when it is a child render object of '
          '[ContentsquareWebViewTrackerBuilder] which reacts to the '
          'hit test',
          (WidgetTester tester) async {
            const key = ValueKey('App');
            await tester.pumpWidget(
              MaterialApp(
                key: key,
                home: ContentsquareWebViewTrackerBuilder(
                  // [GestureDetector] will react to the hit test.
                  builder: (_, __) => GestureDetector(
                    behavior: HitTestBehavior.opaque,
                    child: const SizedBox.expand(),
                  ),
                ),
              ),
            );

            final hitTVPs = getTVPAtHitPosition(
              getCenterFromRenderBoxKey(key: key),
            );

            expect(hitTVPs, isEmpty);
          },
        );

        testWidgets(
          'should not return any TVPs '
          'when all hit elements are children of [ContentsquareWebViewTrackerBuilder]',
          (WidgetTester tester) async {
            const key = ValueKey('App');
            await tester.pumpWidget(
              MaterialApp(
                key: key,
                // [GestureDetector] will react to the hit test but the TVPs
                // to be returned are those of leaf elements, so the one
                // associated with [SizedBox].
                home: GestureDetector(
                  behavior: HitTestBehavior.opaque,
                  child: ContentsquareWebViewTrackerBuilder(
                    builder: (_, __) => const SizedBox.expand(),
                  ),
                ),
              ),
            );

            final hitTVPs = getTVPAtHitPosition(
              getCenterFromRenderBoxKey(key: key),
            );

            expect(hitTVPs, isEmpty);
          },
        );

        testWidgets(
          'should not return any TVPs '
          'when all hit elements are children of [ContentsquareWebViewTrackerBuilder]'
          'and some widgets are placed between a widget that reacts to the hit test '
          'and the [ContentsquareWebViewTrackerBuilder]',
          (WidgetTester tester) async {
            const key = ValueKey('App');
            await tester.pumpWidget(
              MaterialApp(
                key: key,
                // [GestureDetector] will react to the hit test but the TVPs
                // to be returned are those of leaf elements, so the one
                // associated with [SizedBox].
                home: GestureDetector(
                  behavior: HitTestBehavior.opaque,
                  child: SizedBox(
                    child: ContentsquareWebViewTrackerBuilder(
                      builder: (_, __) => const SizedBox.expand(),
                    ),
                  ),
                ),
              ),
            );

            final hitTVPs = getTVPAtHitPosition(
              getCenterFromRenderBoxKey(key: key),
            );

            expect(hitTVPs, isEmpty);
          },
        );

        testWidgets(
          'should only return the TVP from the element outside [ContentsquareWebViewTrackerBuilder] '
          'when the first hit element found is inside [ContentsquareWebViewTrackerBuilder] and the second is outside.',
          (WidgetTester tester) async {
            const key = ValueKey('Stack');
            await tester.pumpWidget(
              MaterialApp(
                home: Scaffold(
                  body: SizedBox(
                    width: 100,
                    height: 100,
                    // The Stack will respond to the hitTest but the [SizedBox]s
                    // will not.
                    child: Stack(
                      key: key,
                      fit: StackFit.expand,
                      children: [
                        // This is in the tapped zone and therefore
                        // should be included but won't be because it is a child
                        // of [ContentsquareWebViewTrackerBuilder].
                        ContentsquareWebViewTrackerBuilder(
                          builder: (_, __) {
                            return const Center(
                              child: SizedBox(height: 50, width: 50),
                            );
                          },
                        ),

                        // This is in the tapped zone and therefore will be
                        // included
                        const Center(
                          child: _NamedWidget<_Included>(
                            child: SizedBox(height: 50, width: 50),
                          ),
                        ),
                      ],
                    ),
                  ),
                ),
              ),
            );

            final hitTVPs = getTVPAtHitPosition(
              getCenterFromRenderBoxKey(key: key),
            );
            final decompressedHitTVPs =
                hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

            expect(decompressedHitTVPs.length, 1);
            expect(
              decompressedHitTVPs.first,
              contains('${typeOf<_NamedWidget<_Included>>()}'),
            );
          },
        );

        testWidgets(
          'should only return the TVP from the element outside [ContentsquareWebViewTrackerBuilder] '
          'when the first hit element found is outside [ContentsquareWebViewTrackerBuilder] and the second is inside.',
          (WidgetTester tester) async {
            const key = ValueKey('Stack');
            await tester.pumpWidget(
              MaterialApp(
                home: Scaffold(
                  body: SizedBox(
                    width: 100,
                    height: 100,
                    // The Stack will respond the the hitTest but the
                    // [SizedBox]s will not.
                    child: Stack(
                      key: key,
                      fit: StackFit.expand,
                      children: [
                        // This is in the tapped zone and therefore will be
                        // included
                        const Center(
                          child: _NamedWidget<_Included>(
                            child: SizedBox(height: 50, width: 50),
                          ),
                        ),

                        // This is in the tapped zone and therefore
                        // should be included but won't be because it is a child
                        // of [ContentsquareWebViewTrackerBuilder].
                        ContentsquareWebViewTrackerBuilder(
                          builder: (_, __) {
                            return const Center(
                              child: SizedBox(height: 50, width: 50),
                            );
                          },
                        ),
                      ],
                    ),
                  ),
                ),
              ),
            );

            final hitTVPs = getTVPAtHitPosition(
              getCenterFromRenderBoxKey(key: key),
            );
            final decompressedHitTVPs =
                hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

            expect(decompressedHitTVPs.length, 1);
            expect(
              decompressedHitTVPs.first,
              contains('${typeOf<_NamedWidget<_Included>>()}'),
            );
          },
        );

        testWidgets(
          'should not return any TVP '
          'when there are multiple webview trackers stacked bellow the widget '
          'which reacts to te hit.',
          (WidgetTester tester) async {
            const key = ValueKey('Stack');
            await tester.pumpWidget(
              MaterialApp(
                home: Scaffold(
                  body: SizedBox(
                    width: 100,
                    height: 100,
                    // The Stack will respond the the hitTest but the
                    // [SizedBox]s will not.
                    child: Stack(
                      key: key,
                      fit: StackFit.expand,
                      children: [
                        // This is in the tapped zone and therefore
                        // should be included but won't be because it is a child
                        // of [ContentsquareWebViewTrackerBuilder].
                        ContentsquareWebViewTrackerBuilder(
                          builder: (_, __) {
                            return const Center(
                              child: SizedBox(height: 50, width: 50),
                            );
                          },
                        ),

                        // This is in the tapped zone and therefore
                        // should be included but won't be because it is a child
                        // of [ContentsquareWebViewTrackerBuilder].
                        ContentsquareWebViewTrackerBuilder(
                          builder: (_, __) {
                            return const Center(
                              child: SizedBox(height: 50, width: 50),
                            );
                          },
                        ),
                      ],
                    ),
                  ),
                ),
              ),
            );

            final hitTVPs = getTVPAtHitPosition(
              getCenterFromRenderBoxKey(key: key),
            );
            final decompressedHitTVPs =
                hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

            expect(decompressedHitTVPs, isEmpty);
          },
        );
      });
    },
  );
}
