import 'package:contentsquare/contentsquare.dart';
import 'package:contentsquare/src/modules/webview_tracking/exceptions/exceptions.dart';
import 'package:contentsquare/src/modules/webview_tracking/logs/logs.dart';
import 'package:contentsquare/src/modules/webview_tracking/screengraph_retrieval.dart';
import 'package:contentsquare/src/modules/webview_tracking/webview_tracking.dart';
import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mocktail/mocktail.dart';
import 'package:webview_flutter/webview_flutter.dart';

import '../../../helpers/decompress_tvp.dart';
import '../../../helpers/expect_logs.dart';
import '../../../helpers/matchers/fixtures/matcher.dart';

class _MockWebViewController extends Mock
    implements ContentsquareWebViewController {}

void main() {
  ContentsquareWebViewTrackerImpl buildDummyWebViewTrackerImplementation() {
    return ContentsquareWebViewTrackerImpl(
      screengraphRetrievalConfiguration: ScreengraphRetrievalConfiguration(
        maxExecutionDuration: const Duration(seconds: 10),
        intervalBetweenJSCalls: const Duration(milliseconds: 100),
      ),
      getEncodedWebViewTrackerTVP: () => null,
    );
  }

  group('$ContentsquareWebViewTrackerBuilder', () {
    testWidgets(
      'should build a WebView tracked by the $ContentsquareWebViewTrackerBuilder',
      (WidgetTester tester) async {
        await tester.pumpWidget(
          MaterialApp(
            home: ContentsquareWebViewTrackerBuilder(
              builder: (context, tracker) {
                return const WebView(
                  initialUrl:
                      'https://tupeuxpastest.csq.io/mobile/web-view-cart.html',
                  javascriptMode: JavascriptMode.unrestricted,
                );
              },
            ),
          ),
        );

        final webview = tester.element(find.byType(WebView));

        expect(webview, isNotNull);
      },
    );

    group('getEncodedWebViewTrackerTVP', () {
      testWidgets(
        'should return the TVP of the WebViewTracker',
        (WidgetTester tester) async {
          late ContentsquareWebViewTrackerImpl trackerImpl;

          await tester.pumpWidget(
            MaterialApp(
              home: ContentsquareWebViewTrackerBuilder(
                builder: (_, tracker) {
                  trackerImpl = (tracker as ContentsquareWebViewTrackerImpl);
                  return const WebView(
                    initialUrl: 'https://www.google.com',
                  );
                },
              ),
            ),
          );

          final tvp = trackerImpl.getEncodedWebViewTrackerTVP();

          expect(decompressTVP(tvp: tvp!), matchesFixture('webviewTVP'));
        },
      );

      testWidgets(
        'should return null and log a $WebViewTrackerNotMountedLog '
        'when the WebViewTracker is not mounted',
        (WidgetTester tester) async {
          late ContentsquareWebViewTrackerImpl trackerImpl;

          await tester.pumpWidget(
            MaterialApp(
              home: Builder(
                builder: (context) {
                  return GestureDetector(
                    behavior: HitTestBehavior.opaque,
                    onTap: () {
                      // Replace the page with a new one to unmount the
                      // ContentsquareWebViewTrackerBuilder
                      Navigator.of(context).pushReplacement(
                        MaterialPageRoute(builder: (_) => const SizedBox()),
                      );
                    },
                    child: ContentsquareWebViewTrackerBuilder(
                      builder: (_, tracker) {
                        trackerImpl =
                            (tracker as ContentsquareWebViewTrackerImpl);
                        return const WebView(
                          initialUrl: 'https://www.google.com',
                        );
                      },
                    ),
                  );
                },
              ),
            ),
          );

          await tester.tap(find.byType(GestureDetector));
          await tester.pumpAndSettle();

          expectLogs((log) => expect(log, isA<WebViewTrackerNotMountedLog>()));

          final webviewTVP = trackerImpl.getEncodedWebViewTrackerTVP();

          expect(webviewTVP, isNull);
        },
      );

      testWidgets(
        'should return null and log a $WebViewTrackerMissingRenderObjectLog '
        'when the flutter trees are only partially built',
        (WidgetTester tester) async {
          late final String? webviewTVP;

          expectLogs(
            // Use [expectSync] because this is triggered inside a [Future]
            // (tester.pumpWidget)
            (log) => expectSync(
              log,
              isA<WebViewTrackerMissingRenderObjectLog>(),
            ),
          );

          await tester.pumpWidget(
            MaterialApp(
              home: ContentsquareWebViewTrackerBuilder(
                builder: (_, tracker) {
                  final trackerImpl =
                      (tracker as ContentsquareWebViewTrackerImpl);

                  // Trigger the error log by calling
                  // getEncodedWebViewTrackerTVP in the middle of the build
                  webviewTVP = trackerImpl.getEncodedWebViewTrackerTVP();
                  return const WebView(initialUrl: 'https://www.google.com');
                },
              ),
            ),
          );

          expect(webviewTVP, isNull);
        },
      );

      testWidgets(
        'should return null and log a $WebViewTrackerMissingGlobalFrameLog '
        'when the widgets have been built but not laid out',
        (WidgetTester tester) async {
          late final String? webviewTVP;

          expectLogs(
            // Use [expectSync] because this is triggered inside a [Future]
            // (tester.pumpWidget)
            (log) => expectSync(
              log,
              isA<WebViewTrackerMissingGlobalFrameLog>(),
            ),
          );

          await tester.pumpWidget(
            MaterialApp(
              home: ContentsquareWebViewTrackerBuilder(
                builder: (_, tracker) {
                  final trackerImpl =
                      (tracker as ContentsquareWebViewTrackerImpl);

                  // LayoutBuilder delays the build of the tree until the widget
                  // above have been laid out
                  return LayoutBuilder(
                    builder: (context, _) {
                      // At this point, the [ContentsquareWebViewRenderObject]
                      // has been built (since it is a parent of the
                      // [LayoutBuilder]) but not laid out (since the build phase
                      // is not over)
                      webviewTVP = trackerImpl.getEncodedWebViewTrackerTVP();
                      return const WebView(
                        initialUrl: 'https://www.google.com',
                      );
                    },
                  );
                },
              ),
            ),
          );

          expect(webviewTVP, isNull);
        },
      );
    });
  });

  group(
    '$ContentsquareWebViewTrackerBuilder injection of the bridge after '
    'onWebViewCreated and onPageFinished have been called',
    () {
      test(
        'should log a $WebViewInitializationErrorLog '
        'when a controller has not been registered',
        () async {
          final tracker = buildDummyWebViewTrackerImplementation();

          expectLogs(
            (log) => expect(log, isA<WebViewInitializationErrorLog>()),
          );

          await tracker.initializeWebViewTracking();
        },
      );

      test(
        'should log a $WebViewInitializationErrorLog '
        'when isChannelAvailable returns false when initializeWebViewTracking is called',
        () async {
          final tracker = buildDummyWebViewTrackerImplementation();

          tracker.registerWebViewController(
            ContentsquareWebViewController(
              evaluateJavascript: (_) => throw UnimplementedError(),
              javascriptChannelCall: (_, __) => throw UnimplementedError(),
              isChannelAvailable: (_) async => false,
            ),
          );

          expectLogs(
            (log) => expect(log, isA<WebViewInitializationErrorLog>()),
          );

          await tracker.initializeWebViewTracking();
        },
      );

      test(
        'should return normally '
        'when the webview controller is registered and isChannelAvailable '
        'returns true when initializeWebViewTracking is called',
        () async {
          final tracker = buildDummyWebViewTrackerImplementation();

          tracker.registerWebViewController(
            ContentsquareWebViewController(
              evaluateJavascript: (_) async => '',
              javascriptChannelCall: (_, __) => '',
              isChannelAvailable: (_) async => true,
            ),
          );

          expectLater(tracker.initializeWebViewTracking, returnsNormally);
        },
      );

      test(
        'The bridge implementation should match its fixture',
        () async {
          late String bridge;
          final tracker = buildDummyWebViewTrackerImplementation();

          tracker.registerWebViewController(
            ContentsquareWebViewController(
              evaluateJavascript: (value) async {
                bridge = value;
                return '';
              },
              javascriptChannelCall: (channelName, data) {
                return '$channelName($data)';
              },
            ),
          );

          expectLater(tracker.initializeWebViewTracking, returnsNormally);

          await expectLater(
            bridge,
            matchesFixture(
              'expectedBridgeJSToDartToInject',
              isVersionSpecific: false,
            ),
          );
        },
      );
    },
  );

  group(
    '$ContentsquareWebViewTrackerImpl',
    () {
      late _MockWebViewController webViewController;

      setUp(() {
        webViewController = _MockWebViewController();
      });

      final defaultConfig = ScreengraphRetrievalConfiguration(
        maxExecutionDuration: const Duration(minutes: 1),
        intervalBetweenJSCalls: const Duration(minutes: 1),
      );
      String? defaultTrackerTVP() => null;

      final customConfig = ScreengraphRetrievalConfiguration(
        maxExecutionDuration: const Duration(days: 1),
        intervalBetweenJSCalls: const Duration(days: 1),
      );
      String? customTrackerTVP() => 'customTrackerTVP';

      ContentsquareWebViewTrackerImpl getTracker({
        ScreengraphRetrievalConfiguration? config,
        String? Function()? trackerTVP,
        bool registerWebViewController = true,
      }) {
        final tracker = ContentsquareWebViewTrackerImpl(
          screengraphRetrievalConfiguration: config ?? defaultConfig,
          getEncodedWebViewTrackerTVP: trackerTVP ?? defaultTrackerTVP,
        );
        if (registerWebViewController) {
          tracker.registerWebViewController(webViewController);
        }
        return tracker;
      }

      test(
        'should be an instance of the $ContentsquareWebViewTracker',
        () {
          final tracker = getTracker();

          expect(tracker, isA<ContentsquareWebViewTracker>());
        },
      );

      test(
        'should have same internal params '
        'when created instance with custom params',
        () {
          final tracker = getTracker(
            config: customConfig,
            trackerTVP: customTrackerTVP,
          );

          expect(
            tracker.screengraphRetrievalConfiguration,
            equals(customConfig),
          );
          expect(
            tracker.getEncodedWebViewTrackerTVP(),
            equals(customTrackerTVP()),
          );
        },
      );

      test(
        'should throw $WebViewNotCreatedScreengraphRetrievalException '
        'when getScreengraph called without registered webViewController',
        () async {
          final tracker = getTracker(registerWebViewController: false);

          expect(
            expectAsync0(tracker.getScreengraph),
            throwsA(isA<WebViewNotCreatedScreengraphRetrievalException>()),
          );
        },
      );

      test(
        'should ${ContentsquareWebViewTrackerImpl.screengraphJavascriptString} '
        'be as expected',
        () async {
          const expectedChannels = '''
    if (window.cs_wvt == undefined || window.cs_wvt == null) {
      // Cast on 'null' for jsonDecode to decode it to dart null.
      'null';
    } else {
      window.cs_wvt.push(['serializeWebView']);
    }
  ''';

          expect(
            ContentsquareWebViewTrackerImpl.screengraphJavascriptString,
            equals(expectedChannels),
          );
        },
      );

      group(
        'initializeWebViewTracking function',
        () {
          test(
            'should NOT initialise when webViewController not registered',
            () async {
              final tracker = getTracker(registerWebViewController: false);
              const expectedErrorText =
                  'WebViewController is null, did you forget to use '
                  'tracker.onWebViewCreated?';
              expectLogs(
                (log) {
                  expect(log, isA<WebViewInitializationErrorLog>());
                  expect(log.message, contains(expectedErrorText));
                },
              );
              await tracker.initializeWebViewTracking();
              verifyNever(() => webViewController.isChannelAvailable);
              verifyZeroInteractions(webViewController);
            },
          );

          test(
            'should log $WebViewInitializationErrorLog '
            'when isChannelAvailable not null but throw error',
            () async {
              final tracker = getTracker();
              final customError = Exception('HI_THERE');
              final expectedErrorText =
                  'Unexpected error while executing JS code, check the '
                  'registration of the javascript channels: $customError';
              when(() => webViewController.isChannelAvailable).thenAnswer(
                (_) => (_) => throw customError,
              );

              expectLogs(
                (log) {
                  expect(log, isA<WebViewInitializationErrorLog>());
                  expect(log.message, contains(expectedErrorText));
                },
              );
              await tracker.initializeWebViewTracking();
              verifyInOrder([
                () => webViewController.isChannelAvailable,
                () => webViewController.currentUrl,
              ]);
              verifyNoMoreInteractions(webViewController);
            },
          );

          test(
            'should log $WebViewInitializationErrorLog '
            'when areJavascriptChannelsRegistered is FALSE',
            () async {
              final tracker = getTracker();
              const expectedErrorText =
                  'JavascriptChannels are not registered, did you forget to '
                  'use tracker.javascriptChannels?';
              when(() => webViewController.isChannelAvailable).thenAnswer(
                (_) => (_) => Future.value(false),
              );

              expectLogs(
                (log) {
                  expect(log, isA<WebViewInitializationErrorLog>());
                  expect(log.message, contains(expectedErrorText));
                },
              );
              await tracker.initializeWebViewTracking();
              verifyInOrder([
                () => webViewController.isChannelAvailable,
                () => webViewController.currentUrl,
              ]);
              verifyNoMoreInteractions(webViewController);
            },
          );

          test(
            'should log $WebViewInitializationErrorLog '
            'when areJavascriptChannelsRegistered is TRUE and evaluateJavascript '
            'throws error',
            () async {
              final tracker = getTracker();
              final customError = Exception('HI_THERE');
              final expectedErrorText =
                  'Unexpected error while executing JS code to setup the '
                  'Javascript bridge: $customError';
              when(() => webViewController.isChannelAvailable).thenAnswer(
                (_) => (_) => Future.value(true),
              );
              when(() => webViewController.evaluateJavascript).thenAnswer(
                (_) => (_) => throw customError,
              );
              when(() => webViewController.javascriptChannelCall)
                  .thenReturn((channelName, data) => '');

              expectLogs(
                (log) {
                  expect(log, isA<WebViewInitializationErrorLog>());
                  expect(log.message, contains(expectedErrorText));
                },
              );
              await tracker.initializeWebViewTracking();
              verifyInOrder([
                () => webViewController.isChannelAvailable,
                () => webViewController.javascriptChannelCall,
                () => webViewController.evaluateJavascript,
                () => webViewController.currentUrl,
              ]);
              verifyNoMoreInteractions(webViewController);
            },
          );

          test(
            'should log $WebViewInitializationSuccessLog '
            'when areJavascriptChannelsRegistered is TRUE and evaluateJavascript '
            'works fine',
            () async {
              final tracker = getTracker();
              const expectedErrorMessage = 'WebView tracking enabled';
              when(() => webViewController.isChannelAvailable).thenAnswer(
                (_) => (_) => Future.value(true),
              );
              when(() => webViewController.evaluateJavascript).thenAnswer(
                (_) => (_) async => '',
              );
              when(() => webViewController.javascriptChannelCall).thenReturn(
                (channelName, data) => '',
              );

              expectLogs(
                (log) {
                  expect(log, isA<WebViewInitializationSuccessLog>());
                  expect(log.message, contains(expectedErrorMessage));
                },
              );
              await tracker.initializeWebViewTracking();
              verifyInOrder([
                () => webViewController.isChannelAvailable,
                () => webViewController.javascriptChannelCall,
                () => webViewController.evaluateJavascript,
                () => webViewController.currentUrl,
              ]);
              verifyNoMoreInteractions(webViewController);
            },
          );
        },
      );
    },
  );
}
